# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox ansible-test sessions.
"""

from __future__ import annotations

import dataclasses
import itertools
import os
import typing as t
from collections.abc import Callable, Sequence
from pathlib import Path

import nox
from antsibull_fileutils.yaml import store_yaml_file

from ..ansible import (
    AnsibleCoreVersion,
    MinPythonVersion,
    MinPythonVersionConstantsWOATC,
    get_ansible_core_info,
    get_ansible_core_package_name,
    get_supported_core_versions,
    parse_ansible_core_version,
)
from ..ansible_test_config import get_min_python_version, load_ansible_test_config
from ..cd import get_base_branch, get_changes, get_vcs_name, is_config_dir_the_repo_dir
from ..config import CONFIG_FILENAME
from ..container import get_container_engine_preference
from ..paths.utils import copy_directory_tree_into
from ..python.versions import get_installed_python_versions
from ..utils import Version
from .collections import prepare_collections
from .utils import (
    normalize_session_name,
    nox_has_color,
    register,
)
from .utils.packages import (
    install,
)
from .utils.values import (
    AnsibleValue,
    AnsibleValueExplicit,
)

if t.TYPE_CHECKING:
    DevelLikeBranch = tuple[str | None, str]


def get_ansible_test_env() -> dict[str, str]:
    """
    Set ANSIBLE_TEST_PREFER_PODMAN if the user prefers one of podman or docker over the other.
    """
    # Check whether the user explicitly set ANSIBLE_TEST_PREFER_PODMAN
    if os.environ.get("ANSIBLE_TEST_PREFER_PODMAN") is not None:
        return {}
    # Check whether the user explicitly requested a container engine for antsibull-nox
    preference, explicitly_set = get_container_engine_preference()
    if not explicitly_set:
        return {}
    # Check whether the users prefers one of podman and docker over the other
    if preference in ("auto-prefer-podman", "podman"):
        # Yes, the user prefers podman.
        return {"ANSIBLE_TEST_PREFER_PODMAN": "1"}
    if preference in ("auto-prefer-docker", "docker"):
        # Yes, the user prefers docker.
        return {"ANSIBLE_TEST_PREFER_PODMAN": ""}
    # Apparently not: do nothing.
    return {}


def get_ansible_test_color_flag(session: nox.Session) -> list[str]:
    """
    Return the fitting --color flag for ansible-test depending on nox' color setting.
    """
    return ["--color", "yes"] if nox_has_color(session) else ["--color", "no"]


class _ColorFlagType:
    pass


COLOR_FLAG = _ColorFlagType()


# NOTE: This is publicly documented API!
# Any change to the API must not be breaking, and must be
# updated in docs/reference.md!
def add_ansible_test_session(
    *,
    name: str,
    description: str | None,
    extra_deps_files: list[str | os.PathLike] | None = None,
    ansible_test_params: Sequence[str | _ColorFlagType],
    add_posargs: bool = True,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    handle_coverage: t.Literal["never", "always", "auto"] = "auto",
    register_name: str | None = None,
    register_extra_data: dict[str, t.Any] | None = None,
    register_tags: Sequence[str] | None = None,
    callback_before: Callable[[], None] | None = None,
    callback_after: Callable[[], None] | None = None,
    support_cd: bool = False,
) -> None:
    """
    Add generic ansible-test session.

    Returns a list of Python versions set for this session.
    """
    parsed_ansible_core_version = parse_ansible_core_version(ansible_core_version)

    def compose_dependencies() -> list[str]:
        deps = [
            get_ansible_core_package_name(
                parsed_ansible_core_version,
                source=ansible_core_source,
                ansible_repo=ansible_core_repo_name,
                branch_name=ansible_core_branch_name,
            )
        ]
        return deps

    def run_ansible_test(session: nox.Session) -> None:
        install(session, *compose_dependencies())

        change_detection_args: list[str] | None = None
        if support_cd and get_vcs_name() == "git" and is_config_dir_the_repo_dir():
            changes = get_changes()
            base_branch = get_base_branch()
            if changes is not None and base_branch is not None:
                if all(file != Path(CONFIG_FILENAME) for file in changes):
                    change_detection_args = [
                        "--changed",
                        "--untracked",
                        "--base-branch",
                        base_branch,
                    ]

        prepared_collections = prepare_collections(
            session,
            ansible_core_version=parsed_ansible_core_version,
            install_in_site_packages=False,
            extra_deps_files=extra_deps_files,
            install_out_of_tree=True,
            copy_repo_structure=change_detection_args is not None,
        )
        if not prepared_collections:
            session.warn("Skipping ansible-test...")
            return
        cwd = Path.cwd()
        with session.chdir(prepared_collections.current_path):
            if callback_before:
                callback_before()

            command = ["ansible-test"]
            for param in ansible_test_params:
                if isinstance(param, _ColorFlagType):
                    command.extend(get_ansible_test_color_flag(session))
                else:
                    command.append(param)
            if change_detection_args is not None:
                command.extend(change_detection_args)
            if add_posargs and session.posargs:
                command.extend(session.posargs)
            session.run(*command, env=get_ansible_test_env())

            coverage = (handle_coverage == "auto" and "--coverage" in command) or (
                handle_coverage == "always"
            )
            if coverage:
                session.run(
                    "ansible-test",
                    "coverage",
                    "xml",
                    *get_ansible_test_color_flag(session),
                    "-v",
                    "--requirements",
                    "--group-by",
                    "command",
                    "--group-by",
                    "version",
                )

            if callback_after:
                callback_after()

            copy_directory_tree_into(
                prepared_collections.current_path / "tests" / "output",
                cwd / "tests" / "output",
            )

    # Determine Python version(s)
    core_info = get_ansible_core_info(parsed_ansible_core_version)
    all_versions = get_installed_python_versions()

    installed_versions = [
        version
        for version in core_info.controller_python_versions
        if version in all_versions
    ]
    python = max(installed_versions or core_info.controller_python_versions)
    python_versions = [python]

    run_ansible_test.__doc__ = description
    nox.session(
        name=name,
        default=default,
        python=[str(python_version) for python_version in python_versions],
    )(run_ansible_test)

    if register_name:
        data: dict[str, t.Any] = {
            "name": name,
            "ansible-core": (
                str(ansible_core_branch_name)
                if ansible_core_branch_name is not None
                else str(parsed_ansible_core_version)
            ),
            "python": " ".join(str(python) for python in python_versions),
        }
        if register_extra_data:
            data.update(register_extra_data)
        if register_tags:
            data["tags"] = sorted(register_tags)
        register(register_name, data)


def add_ansible_test_sanity_test_session(
    *,
    name: str,
    description: str | None,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    skip_tests: list[str] | None = None,
    allow_disabled: bool = False,
    enable_optional_errors: bool = False,
    register_extra_data: dict[str, t.Any] | None = None,
) -> None:
    """
    Add generic ansible-test sanity test session.
    """
    command: list[str | _ColorFlagType] = ["sanity", COLOR_FLAG, "-v", "--docker"]
    if skip_tests:
        for test in skip_tests:
            command.extend(["--skip", test])
    if allow_disabled:
        command.append("--allow-disabled")
    if enable_optional_errors:
        command.append("--enable-optional-errors")
    add_ansible_test_session(
        name=name,
        description=description,
        ansible_test_params=command,
        default=default,
        ansible_core_version=ansible_core_version,
        ansible_core_source=ansible_core_source,
        ansible_core_repo_name=ansible_core_repo_name,
        ansible_core_branch_name=ansible_core_branch_name,
        register_name="sanity",
        register_extra_data=register_extra_data,
        register_tags=["sanity", "docker"],
        support_cd=True,
    )


def _parse_min_max_except(
    min_version: Version | str | None,
    max_version: Version | str | None,
    except_versions: list[AnsibleCoreVersion | str] | None,
) -> tuple[Version | None, Version | None, tuple[AnsibleCoreVersion, ...] | None]:
    if isinstance(min_version, str):
        min_version = Version.parse(min_version)
    if isinstance(max_version, str):
        max_version = Version.parse(max_version)
    if except_versions is None:
        return min_version, max_version, None
    evs = tuple(parse_ansible_core_version(version) for version in except_versions)
    return min_version, max_version, evs


def add_all_ansible_test_sanity_test_sessions(
    *,
    default: bool = False,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[DevelLikeBranch] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
    skip_tests: list[str] | None = None,
    allow_disabled: bool = False,
    enable_optional_errors: bool = False,
) -> None:
    """
    Add ansible-test sanity test meta session that runs ansible-test sanity
    for all supported ansible-core versions.
    """
    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )

    sanity_sessions = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        name = f"ansible-test-sanity-{ansible_core_version}"
        add_ansible_test_sanity_test_session(
            name=name,
            description=f"Run sanity tests from ansible-core {ansible_core_version}'s ansible-test",
            ansible_core_version=ansible_core_version,
            skip_tests=skip_tests,
            allow_disabled=allow_disabled,
            enable_optional_errors=enable_optional_errors,
            register_extra_data={"display-name": f"Ⓐ{ansible_core_version}"},
            default=False,
        )
        sanity_sessions.append(name)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            repo_postfix = f", {repo_name} repository" if repo_name is not None else ""
            name = f"ansible-test-sanity-{repo_prefix}{branch_name.replace('/', '-')}"
            add_ansible_test_sanity_test_session(
                name=name,
                description=(
                    "Run sanity tests from ansible-test in ansible-core's"
                    f" {branch_name} branch{repo_postfix}"
                ),
                ansible_core_version="devel",
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                skip_tests=skip_tests,
                allow_disabled=allow_disabled,
                enable_optional_errors=enable_optional_errors,
                register_extra_data={"display-name": f"Ⓐ{branch_name}"},
                default=False,
            )
            sanity_sessions.append(name)

    def run_all_sanity_tests(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    run_all_sanity_tests.__doc__ = (
        "Meta session for running all ansible-test-sanity-* sessions."
    )
    nox.session(
        name="ansible-test-sanity",
        default=default,
        requires=sanity_sessions,
    )(run_all_sanity_tests)


def add_ansible_test_unit_test_session(
    *,
    name: str,
    description: str | None,
    default: bool,
    ansible_core_version: str | AnsibleCoreVersion,
    ansible_core_source: t.Literal["git", "pypi"] = "git",
    ansible_core_repo_name: str | None = None,
    ansible_core_branch_name: str | None = None,
    register_extra_data: dict[str, t.Any] | None = None,
) -> None:
    """
    Add generic ansible-test unit test session.
    """
    add_ansible_test_session(
        name=name,
        description=description,
        ansible_test_params=["units", COLOR_FLAG, "-v", "--docker"],
        extra_deps_files=["tests/unit/requirements.yml"],
        default=default,
        ansible_core_version=ansible_core_version,
        ansible_core_source=ansible_core_source,
        ansible_core_repo_name=ansible_core_repo_name,
        ansible_core_branch_name=ansible_core_branch_name,
        register_name="units",
        register_extra_data=register_extra_data,
        register_tags=["units", "docker"],
        support_cd=True,
    )


def add_all_ansible_test_unit_test_sessions(
    *,
    default: bool = False,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[DevelLikeBranch] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
) -> None:
    """
    Add ansible-test unit test meta session that runs ansible-test units
    for all supported ansible-core versions.
    """
    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )

    units_sessions = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        name = f"ansible-test-units-{ansible_core_version}"
        add_ansible_test_unit_test_session(
            name=name,
            description=f"Run unit tests with ansible-core {ansible_core_version}'s ansible-test",
            ansible_core_version=ansible_core_version,
            register_extra_data={"display-name": f"Ⓐ{ansible_core_version}"},
            default=False,
        )
        units_sessions.append(name)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            repo_postfix = f", {repo_name} repository" if repo_name is not None else ""
            name = f"ansible-test-units-{repo_prefix}{branch_name.replace('/', '-')}"
            add_ansible_test_unit_test_session(
                name=name,
                description=(
                    "Run unit tests from ansible-test in ansible-core's"
                    f" {branch_name} branch{repo_postfix}"
                ),
                ansible_core_version="devel",
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                register_extra_data={"display-name": f"Ⓐ{branch_name}"},
                default=False,
            )
            units_sessions.append(name)

    def run_all_unit_tests(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        pass

    run_all_unit_tests.__doc__ = (
        "Meta session for running all ansible-test-units-* sessions."
    )
    nox.session(
        name="ansible-test-units",
        default=default,
        requires=units_sessions,
    )(run_all_unit_tests)


def update_gitignore(gitignore_path: Path, *paths_to_add: Path) -> None:
    """
    Add the given paths to .gitignore.
    """
    if not paths_to_add:
        return

    existing: list[str]
    try:
        with open(gitignore_path, encoding="utf-8", mode="rt") as f:
            existing = [f.read()]
    except FileNotFoundError:
        existing = [""]

    for path in paths_to_add:
        existing.append(f"/{path}")
    existing.append("")

    with open(gitignore_path, encoding="utf-8", mode="wt") as f:
        f.write("\n".join(existing))


def write_integration_config(
    *ansible_vars: dict[str, AnsibleValue] | None,
    ansible_vars_from_env_vars: dict[str, str] | None = None,
) -> None:
    """
    Write tests/integration/integration_config.yml with the provided values.
    """
    path = Path("tests", "integration", "integration_config.yml")
    content: dict[str, t.Any] = {}

    if ansible_vars_from_env_vars:
        for ans_var, env_var in ansible_vars_from_env_vars.items():
            value = os.environ.get(env_var)
            if value is not None:
                content[ans_var] = env_var

    for one_ansible_vars in ansible_vars:
        if one_ansible_vars:
            for ans_var, ans_value in one_ansible_vars.items():
                val, store = ans_value.get_value()
                if store:
                    content[ans_var] = val

    store_yaml_file(path, content, nice=True, sort_keys=True)

    update_gitignore(
        Path(".gitignore"),
        path,
    )


def _add_version_specific_interation_test_params(
    parameters: list[str | _ColorFlagType], ansible_core_version: AnsibleCoreVersion
) -> None:
    version = get_ansible_core_info(ansible_core_version).ansible_core_version
    if version >= Version(2, 19):
        # The behavior changed from ansible-core 2.19 on, that tracebacks are no
        # longer shown in case of errors by default. While for end-users this is
        # OK, this is potentially really bad for integration tests. In case of
        # *intentional* errors this is also fine (and reduces clutter), but it
        # also prevents stack traces being shown in case of *unintentional* errors.
        # So we generally enable tracebacks for errors for integration tests
        # since this seems like the lesser evil.
        # (Related: https://github.com/ansible/ansible/issues/86533)
        parameters.extend(["--display-traceback", "error"])


def add_ansible_test_integration_sessions_default_container(
    *,
    include_devel: bool = False,
    include_milestone: bool = False,
    add_devel_like_branches: list[DevelLikeBranch] | None = None,
    min_version: Version | str | None = None,
    max_version: Version | str | None = None,
    except_versions: list[AnsibleCoreVersion | str] | None = None,
    core_python_versions: (
        dict[str | AnsibleCoreVersion, list[str | Version]] | None
    ) = None,
    controller_python_versions_only: bool = False,
    min_python_version: MinPythonVersion | None = None,
    ansible_vars_from_env_vars: dict[str, str] | None = None,
    ansible_vars: dict[str, AnsibleValue] | None = None,
    default: bool = False,
) -> list[str]:
    """
    Add ansible-test integration tests using the default Docker container.

    ``core_python_versions`` can be used to restrict the Python versions
    to be used for a specific ansible-core version.

    ``controller_python_versions_only`` can be used to only run against
    controller Python versions.

    ``min_python_version`` allows to generally restrict the Python versions.
    This is only used when ``core_python_versions`` does not provide an
    explicit list of Python versions.
    """
    effective_min_python_version: (
        MinPythonVersionConstantsWOATC | Version | t.Callable[[Version], bool]
    )
    if min_python_version == "ansible-test-config":
        effective_min_python_version = get_min_python_version(
            load_ansible_test_config(ignore_errors=True), ignore_errors=True
        )
    else:
        effective_min_python_version = min_python_version or "default"

    def callback_before() -> None:
        if not ansible_vars_from_env_vars and not ansible_vars:
            return

        write_integration_config(
            ansible_vars, ansible_vars_from_env_vars=ansible_vars_from_env_vars
        )

    def add_integration_tests(
        ansible_core_version: AnsibleCoreVersion,
        repo_name: str | None = None,
        branch_name: str | None = None,
    ) -> list[str]:
        # Determine Python versions to run tests for
        py_versions = (
            (core_python_versions.get(branch_name) if branch_name is not None else None)
            or core_python_versions.get(ansible_core_version)
            or core_python_versions.get(str(ansible_core_version))
            if core_python_versions
            else None
        )
        if py_versions is None:
            core_info = get_ansible_core_info(ansible_core_version)
            effective_py_versions: list[Version] = list(
                core_info.controller_python_versions
                if controller_python_versions_only
                or effective_min_python_version == "controller"
                else core_info.remote_python_versions
            )
            if effective_min_python_version not in ("default", "controller"):
                if callable(effective_min_python_version):
                    # effective_min_python_version is a predicate
                    effective_py_versions = [
                        pyv
                        for pyv in effective_py_versions
                        if effective_min_python_version(pyv)
                    ]
                else:
                    # mypy doesn't get this, so we have to use assert()...
                    assert isinstance(effective_min_python_version, Version)
                    effective_py_versions = [
                        pyv
                        for pyv in effective_py_versions
                        if pyv >= effective_min_python_version
                    ]
            py_versions = list(effective_py_versions)

        # Add sessions
        integration_sessions_core: list[str] = []
        if branch_name is None:
            base_name = f"ansible-test-integration-{ansible_core_version}-"
        else:
            repo_prefix = (
                f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
            )
            base_name = f"ansible-test-integration-{repo_prefix}{branch_name.replace('/', '-')}-"
        for py_version in py_versions:
            name = f"{base_name}{py_version}"
            if branch_name is None:
                description = (
                    f"Run integration tests from ansible-core {ansible_core_version}'s"
                    f" ansible-test with Python {py_version}"
                )
            else:
                repo_postfix = (
                    f", {repo_name} repository" if repo_name is not None else ""
                )
                description = (
                    f"Run integration tests from ansible-test in ansible-core's {branch_name}"
                    f" branch{repo_postfix} with Python {py_version}"
                )
            parameters: list[str | _ColorFlagType] = [
                "integration",
                COLOR_FLAG,
                "-v",
                "--docker",
                "default",
                "--python",
                str(py_version),
            ]
            _add_version_specific_interation_test_params(
                parameters, ansible_core_version
            )
            add_ansible_test_session(
                name=name,
                description=description,
                ansible_test_params=parameters,
                extra_deps_files=["tests/integration/requirements.yml"],
                ansible_core_version=ansible_core_version,
                ansible_core_repo_name=repo_name,
                ansible_core_branch_name=branch_name,
                callback_before=callback_before,
                default=False,
                register_name="integration",
                register_extra_data={
                    "display-name": f"Ⓐ{ansible_core_version}+py{py_version}+default"
                },
                register_tags=["integration", "docker", "docker-default"],
                support_cd=True,
            )
            integration_sessions_core.append(name)
        return integration_sessions_core

    parsed_min_version, parsed_max_version, parsed_except_versions = (
        _parse_min_max_except(min_version, max_version, except_versions)
    )
    integration_sessions: list[str] = []
    for ansible_core_version in get_supported_core_versions(
        include_devel=include_devel,
        include_milestone=include_milestone,
        min_version=parsed_min_version,
        max_version=parsed_max_version,
        except_versions=parsed_except_versions,
    ):
        integration_sessions_core = add_integration_tests(ansible_core_version)
        if integration_sessions_core:
            name = f"ansible-test-integration-{ansible_core_version}"
            integration_sessions.append(name)

            def run_integration_tests(
                session: nox.Session,  # pylint: disable=unused-argument
            ) -> None:
                pass

            run_integration_tests.__doc__ = (
                f"Meta session for running all {name}-* sessions."
            )
            nox.session(
                name=name,
                requires=integration_sessions_core,
                default=default,
            )(run_integration_tests)
    if add_devel_like_branches:
        for repo_name, branch_name in add_devel_like_branches:
            integration_sessions_core = add_integration_tests(
                "devel", repo_name=repo_name, branch_name=branch_name
            )
            if integration_sessions_core:
                repo_prefix = (
                    f"{repo_name.replace('/', '-')}-" if repo_name is not None else ""
                )
                name = f"ansible-test-integration-{repo_prefix}{branch_name.replace('/', '-')}"
                integration_sessions.append(name)

                def run_integration_tests_for_branch(
                    session: nox.Session,  # pylint: disable=unused-argument
                ) -> None:
                    pass

                run_integration_tests_for_branch.__doc__ = (
                    f"Meta session for running all {name}-* sessions."
                )
                nox.session(
                    name=name,
                    requires=integration_sessions_core,
                    default=default,
                )(run_integration_tests_for_branch)

    return integration_sessions


@dataclasses.dataclass
class AnsibleTestIntegrationSessionTemplate:
    """
    A template for an ansible-test integration test session.
    """

    ansible_core: list[AnsibleCoreVersion]
    docker: list[str | None]
    remote: list[str | None]
    python_version: list[Version | None]
    target: list[str | None]
    gha_container: list[str | None]

    devel_like_branch: DevelLikeBranch | None

    ansible_vars: dict[str, AnsibleValue]
    session_name_template: str
    display_name_template: str
    description_template: str
    tags: list[str]


@dataclasses.dataclass
class AnsibleTestIntegrationSessionTemplateGroup:
    """
    A template for an ansible-test integration test session.
    """

    session_name: str | None
    description: str | None
    ansible_vars: dict[str, AnsibleValue]
    session_templates: list[AnsibleTestIntegrationSessionTemplate]
    tags: list[str]


@dataclasses.dataclass
class AnsibleTestIntegrationSession:
    """
    A template for an ansible-test integration test session.
    """

    source: str
    part_of_group: bool

    ansible_core: AnsibleCoreVersion
    ansible_core_repo_name: str | None
    ansible_core_branch_name: str | None
    docker: str | None
    remote: str | None
    python_version: Version | None
    target: str | None
    gha_container: str | None

    ansible_vars: dict[str, AnsibleValue]
    session_name: str
    display_name: str
    description: str
    tags: list[str]

    def get_ansible_vars_callback(self) -> t.Callable[[], None] | None:
        """
        Create before-callback for ansible-test.
        """
        if not self.ansible_vars:
            return None

        def callback_before() -> None:
            write_integration_config(self.ansible_vars)

        return callback_before


@dataclasses.dataclass
class AnsibleTestIntegrationSessionGroup:
    """
    A template for an ansible-test integration test session.
    """

    session_name: str | None
    description: str | None
    sessions: list[AnsibleTestIntegrationSession]


def _get_templator(**kwargs: t.Any) -> t.Callable[[str], str]:
    template_vars = {}
    for k, v in kwargs.items():
        val = str(v) if v is not None else ""
        template_vars[k] = val
        template_vars[f"{k}_dash"] = f"{val}-" if val else ""
        template_vars[f"dash_{k}"] = f"-{val}" if val else ""
        template_vars[f"{k}_plus"] = f"{val}+" if val else ""
        template_vars[f"plus_{k}"] = f"+{val}" if val else ""
        template_vars[f"{k}_comma"] = f"{val}, " if val else ""
        template_vars[f"comma_{k}"] = f", {val}" if val else ""

    def tmpl(template: str) -> str:
        try:
            return template.format(**template_vars)
        except KeyError as exc:
            raise ValueError(
                f"Error when templating {template!r}: unknown variable {str(exc)!r}"
            ) from None

    return tmpl


def _template_session(
    session_template: AnsibleTestIntegrationSessionTemplate,
    source: str,
    part_of_group: bool,
    ansible_vars: list[dict[str, AnsibleValue] | None],
    tags: set[str] | list[str],
) -> t.Generator[AnsibleTestIntegrationSession]:
    session_ansible_vars = {}
    for ansible_vars_item in ansible_vars:
        if ansible_vars_item:
            session_ansible_vars.update(ansible_vars_item)
    session_ansible_vars.update(session_template.ansible_vars)
    vars_values: dict[str, t.Any] = {}
    for var, value in session_ansible_vars.items():
        if value.template_value is not None:
            vars_values[var] = value.template_value
        elif isinstance(value, AnsibleValueExplicit):
            vars_values[var] = value.value
    session_tags = set(tags)
    session_tags.update(session_template.tags)
    tags_list = sorted(session_tags)
    for (
        ansible_core,
        docker,
        remote,
        python_version,
        target,
        gha_container,
    ) in itertools.product(
        session_template.ansible_core,
        session_template.docker,
        session_template.remote,
        session_template.python_version,
        session_template.target,
        session_template.gha_container,
    ):
        gha_arm = gha_container and "-arm" in gha_container
        vars_values.update(
            {
                "ansible_core": ansible_core,
                "docker": docker,
                "docker_short": (
                    (
                        docker.removeprefix(
                            "quay.io/ansible-community/test-image:"
                        ).removeprefix("localhost/test-image:")
                    )
                    if docker
                    else None
                ),
                "remote": remote,
                "python_version": python_version,
                "py_python_version": f"py{python_version}" if python_version else None,
                "target": target,
                "target_dashized": (
                    target.replace("/", "-").strip("-") if target else None
                ),
                "gha_container": gha_container,
                "gha_arm": "ARM" if gha_arm else None,
                "gha_arm_lower": "arm" if gha_arm else None,
            }
        )
        tmpl = _get_templator(**vars_values)
        yield AnsibleTestIntegrationSession(
            source=source,
            part_of_group=part_of_group,
            ansible_core=ansible_core,
            ansible_core_repo_name=(
                session_template.devel_like_branch[0]
                if ansible_core == "devel" and session_template.devel_like_branch
                else None
            ),
            ansible_core_branch_name=(
                session_template.devel_like_branch[1]
                if ansible_core == "devel" and session_template.devel_like_branch
                else None
            ),
            docker=docker,
            remote=remote,
            python_version=python_version,
            target=target,
            gha_container=gha_container,
            ansible_vars=session_ansible_vars,
            session_name=normalize_session_name(
                tmpl(session_template.session_name_template)
            ),
            display_name=tmpl(session_template.display_name_template),
            description=tmpl(session_template.description_template),
            tags=tags_list,
        )


def _template_sessions(
    session_templates: list[AnsibleTestIntegrationSessionTemplate],
    session_template_groups: list[AnsibleTestIntegrationSessionTemplateGroup],
    ansible_vars: dict[str, AnsibleValue],
    tags: list[str],
) -> tuple[
    list[AnsibleTestIntegrationSession], list[AnsibleTestIntegrationSessionGroup]
]:
    result: list[AnsibleTestIntegrationSession] = []
    result_groups: list[AnsibleTestIntegrationSessionGroup] = []
    for index, template in enumerate(session_templates):
        for session in _template_session(
            template, f"session template #{index + 1}", False, [ansible_vars], tags
        ):
            result.append(session)
    for group_index, group in enumerate(session_template_groups):
        group_sessions: list[AnsibleTestIntegrationSession] = []
        group_tags = set(tags)
        group_tags.update(group.tags)
        for index, template in enumerate(group.session_templates):
            for session in _template_session(
                template,
                f"session template #{index + 1} of group #{group_index + 1}",
                True,
                [ansible_vars, group.ansible_vars],
                group_tags,
            ):
                result.append(session)
                group_sessions.append(session)
        result_groups.append(
            AnsibleTestIntegrationSessionGroup(
                session_name=group.session_name,
                description=group.description,
                sessions=group_sessions,
            )
        )
    return result, result_groups


def add_ansible_test_integration_sessions(
    *,
    session_templates: list[AnsibleTestIntegrationSessionTemplate] | None = None,
    session_template_groups: (
        list[AnsibleTestIntegrationSessionTemplateGroup] | None
    ) = None,
    ansible_vars: dict[str, AnsibleValue] | None = None,
    global_tags: list[str] | None = None,
    default: bool = False,
) -> list[str]:
    """
    Add ansible-test integration tests.
    """
    sessions, groups = _template_sessions(
        session_templates or [],
        session_template_groups or [],
        ansible_vars or {},
        global_tags or [],
    )
    session_by_name: dict[str, AnsibleTestIntegrationSession] = {}
    for session in sessions:
        # nox does not warn/error on duplicate session names
        # (https://github.com/wntrblm/nox/issues/998)
        if session.session_name in session_by_name:
            raise ValueError(
                f"Collision: {session.source} and {session_by_name[session.session_name].source}"
                f" both generated the same session name {session.session_name!r}"
            )
        session_by_name[session.session_name] = session

    for _, session in sorted(session_by_name.items()):
        register_tags = ["integration"] + session.tags
        cmd: list[str | _ColorFlagType] = [
            "integration",
            COLOR_FLAG,
            "-v",
        ]
        _add_version_specific_interation_test_params(cmd, session.ansible_core)
        if session.docker:
            cmd.extend(
                [
                    "--docker",
                    session.docker,
                ]
            )
            register_tags.extend(["docker", f"docker-{session.docker}"])
        if session.remote:
            cmd.extend(
                [
                    "--remote",
                    session.remote,
                ]
            )
            register_tags.extend(["remote", f"remote-{session.remote}"])
        if session.python_version:
            cmd.extend(
                [
                    "--python",
                    str(session.python_version),
                ]
            )
            register_tags.extend(["python", f"python-{session.python_version}"])
        if session.target:
            cmd.append(session.target)
        extra_data = {
            "display-name": session.display_name,
        }
        if session.gha_container is not None:
            extra_data["gha-container"] = session.gha_container
        add_ansible_test_session(
            name=session.session_name,
            description=session.description,
            ansible_test_params=cmd,
            extra_deps_files=["tests/integration/requirements.yml"],
            ansible_core_version=session.ansible_core,
            ansible_core_repo_name=session.ansible_core_repo_name,
            ansible_core_branch_name=session.ansible_core_repo_name,
            callback_before=session.get_ansible_vars_callback(),
            default=default and not session.part_of_group,
            register_name="integration",
            register_extra_data=extra_data,
            register_tags=sorted(register_tags),
            support_cd=True,
        )

    def add_group_session(
        session_name: str, description: str, session_names: list[str]
    ) -> None:
        def ansible_test_group(
            session: nox.Session,  # pylint: disable=unused-argument
        ) -> None:
            pass

        ansible_test_group.__doc__ = description
        nox.session(
            name=session_name,
            requires=sorted(session_names),
            default=default,
        )(ansible_test_group)

    for group in groups:
        if group.session_name is None:
            continue
        add_group_session(
            normalize_session_name(group.session_name),
            group.description
            or "Meta session for running some ansible-test integration test sessions.",
            [session.session_name for session in group.sessions],
        )

    return sorted(session_by_name)


__all__ = [
    "AnsibleTestIntegrationSessionTemplate",
    "AnsibleTestIntegrationSessionTemplateGroup",
    "add_ansible_test_session",
    "add_ansible_test_sanity_test_session",
    "add_all_ansible_test_sanity_test_sessions",
    "add_ansible_test_unit_test_session",
    "add_all_ansible_test_unit_test_sessions",
    "add_ansible_test_integration_sessions_default_container",
    "add_ansible_test_integration_sessions",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for handling scripts.
"""

from __future__ import annotations

import os
import sys
import typing as t
from pathlib import Path

import nox

from ...data_util import prepare_data_script
from ...messages.parse import parse_bare_framework_errors
from ...paths.utils import (
    find_data_directory,
    list_all_files,
)
from ..utils.output import print_messages
from . import silence_run_verbosity
from .paths import filter_files_cd


def run_bare_script(
    session: nox.Session,
    /,
    name: str,
    *,
    use_session_python: bool = False,
    files: list[Path] | None = None,
    extra_data: dict[str, t.Any] | None = None,
    silent: bool = False,
    with_cd: bool = False,
    process_messages: bool = False,
) -> str | None:
    """
    Run a bare script included in antsibull-nox's data directory.
    """
    if files is None:
        files = list_all_files()
    if with_cd:
        files = filter_files_cd(files)
        if not files:
            session.warn(f"Skipping {name} (no files to process)")
            return None
    data = prepare_data_script(
        session,
        base_name=name,
        paths=files,
        extra_data=extra_data,
    )
    python = sys.executable
    env = {}
    if use_session_python:
        python = "python"
        env["PYTHONPATH"] = str(find_data_directory())
    command: list[str | os.PathLike[str]] = [
        python,
        find_data_directory() / f"{name}.py",
        "--data",
        data,
    ]
    kwargs: dict[str, t.Any] = {
        "external": True,
        "silent": silent,
        "env": env,
    }

    if process_messages:
        kwargs["silent"] = True
        kwargs["success_codes"] = (0, 1)
        with silence_run_verbosity():
            output = session.run(*command, **kwargs)

        if output:
            print_messages(
                session=session,
                messages=parse_bare_framework_errors(
                    output=output,
                ),
                fail_msg=f"{name} failed",
            )

    else:
        output = session.run(*command, **kwargs)

    return output


__all__ = [
    "run_bare_script",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

"""
Tests for the collection module
"""

from __future__ import annotations

import typing as t
from pathlib import Path

import pytest
from antsibull_fileutils.yaml import load_yaml_file

from antsibull_nox.collection.utils import (
    force_collection_version,
)

FORCE_COLLECTION_VERSION_DATA: list[tuple[str, str, bool, dict[str, t.Any]]] = [
    (
        r"""name: foo""",
        "1.2.3",
        True,
        {
            "name": "foo",
            "version": "1.2.3",
        },
    ),
    (
        r"""version: []""",
        "1.2.3",
        True,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""version: 1.2.2""",
        "1.2.3",
        True,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""version: 1.2.3""",
        "1.2.3",
        False,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""
name: boo
namespace: foo
version: null""",
        "1.2.3",
        True,
        {
            "name": "boo",
            "namespace": "foo",
            "version": "1.2.3",
        },
    ),
]


@pytest.mark.parametrize(
    "content, version, expected_result, expected_content",
    FORCE_COLLECTION_VERSION_DATA,
)
def test_force_collection_version(
    content: str,
    version: str,
    expected_result: bool,
    expected_content: dict[str, t.Any],
    tmp_path: Path,
) -> None:
    file = tmp_path / "galaxy.yml"
    file.write_text(content)
    result = force_collection_version(tmp_path, version=version)
    assert result == expected_result
    assert load_yaml_file(file) == expected_content


FORCE_COLLECTION_VERSION_FAIL_DATA: list[tuple[str, str, str]] = [
    (
        r"""{""",
        "1.2.3",
        "^Cannot parse .*/galaxy.yml: ",
    ),
]


@pytest.mark.parametrize(
    "content, version, expected_match",
    FORCE_COLLECTION_VERSION_FAIL_DATA,
)
def test_force_collection_version_fail(
    content: str, version: str, expected_match: str, tmp_path: Path
) -> None:
    file = tmp_path / "galaxy.yml"
    file.write_text(content)
    with pytest.raises(ValueError, match=expected_match):
        force_collection_version(tmp_path, version=version)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Config file schema.
"""

from __future__ import annotations

import os
import typing as t

import pydantic as p

from antsibull_nox.ee_config import create_ee_config

from ._pydantic import forbid_extras, get_formatted_error_messages
from .ansible import AnsibleCoreVersion, MinPythonVersionConstants
from .data.antsibull_nox_data_util import Message
from .sessions.utils.packages import PackageConstraints as _PackageConstraints
from .sessions.utils.packages import PackageEditable as _PackageEditable
from .sessions.utils.packages import PackageName as _PackageName
from .sessions.utils.packages import PackageRequirements as _PackageRequirements
from .sessions.utils.values import AnsibleValueExplicit as _AnsibleValueExplicit
from .sessions.utils.values import AnsibleValueFromEnv as _AnsibleValueFromEnv
from .utils import Version

try:
    from tomllib import load as _load_toml
except ImportError:
    from tomli import load as _load_toml  # type: ignore


CONFIG_FILENAME = "antsibull-nox.toml"


def _parse_version(value: t.Any) -> Version:
    if isinstance(value, Version):
        return value
    if isinstance(value, str) and "." in value:
        return Version.parse(value)
    raise ValueError("Must be version string")


def _parse_ansible_core_version(value: t.Any) -> AnsibleCoreVersion:
    if isinstance(value, Version):
        return value
    if isinstance(value, str):
        if value == "devel":
            return "devel"
        if value == "milestone":
            return "milestone"
        if "." in value:
            return Version.parse(value)
    raise ValueError("Must be ansible-core version string")


def _validate_collection_name(value: str) -> str:
    parts = value.split(".")
    if len(parts) != 2:
        raise ValueError("Collection name must be of the form '<namespace>.<name>'")
    if not parts[0].isidentifier():
        raise ValueError("Collection namespace must be Python identifier")
    if not parts[1].isidentifier():
        raise ValueError("Collection name must be Python identifier")
    return value


CollectionName = t.Annotated[str, p.AfterValidator(_validate_collection_name)]
PVersion = t.Annotated[Version, p.BeforeValidator(_parse_version)]
PAnsibleCoreVersion = t.Annotated[
    AnsibleCoreVersion, p.BeforeValidator(_parse_ansible_core_version)
]


class PackageName(p.BaseModel):
    """
    A PyPI package name.
    """

    type: t.Literal["package"] = "package"
    name: str

    def to_utils_instance(self) -> _PackageName:
        """
        Convert config object to runtime package name object.
        """
        return _PackageName(name=self.name)


class PackageEditable(p.BaseModel):
    """
    A PyPI package name that should be installed editably (if allowed).
    """

    type: t.Literal["editable"] = "editable"
    name: str

    def to_utils_instance(self) -> _PackageEditable:
        """
        Convert config object to runtime editable package name object.
        """
        return _PackageEditable(name=self.name)


class PackageRequirements(p.BaseModel):
    """
    A Python requirements.txt file.
    """

    type: t.Literal["requirements"] = "requirements"
    file: str

    def to_utils_instance(self) -> _PackageRequirements:
        """
        Convert config object to runtime package requirements object.
        """
        return _PackageRequirements(file=self.file)


class PackageConstraints(p.BaseModel):
    """
    A Python constraints.txt file.
    """

    type: t.Literal["constraints"] = "constraints"
    file: str

    def to_utils_instance(self) -> _PackageConstraints:
        """
        Convert config object to runtime package requirements object.
        """
        return _PackageConstraints(file=self.file)


PackageType = t.Union[
    PackageName,
    PackageEditable,
    PackageRequirements,
    PackageConstraints,
]


_VALID_PACKAGE_TYPE_NAMES = tuple(
    p.model_fields["type"].default for p in t.get_args(PackageType)
)


def _package_name_validator(value: t.Any, *, keep_strings: bool = False) -> t.Any:
    """
    Convert plain strings into PackageName instances.
    """
    if isinstance(value, str):
        if keep_strings:
            return value
        return PackageName(name=value)
    if isinstance(value, dict):
        # Special-casing for "type" to provide a clean error message
        if "type" not in value or value["type"] not in _VALID_PACKAGE_TYPE_NAMES:
            raise ValueError(
                "Must provide a valid 'type' when specifying a package."
                + f" Valid types are: {', '.join(_VALID_PACKAGE_TYPE_NAMES)}"
            )
    return value


PackageField = t.Annotated[
    PackageType,
    p.Field(discriminator="type"),
    p.BeforeValidator(_package_name_validator),
]


def _package_string_discriminator(value: t.Any) -> t.Literal["package_type", "string"]:
    """
    Discriminate between a string and a package type.
    """
    if isinstance(value, str):
        return "string"
    return "package_type"


PackageFieldOrString = t.Annotated[
    t.Union[
        t.Annotated[PackageField, p.Tag("package_type")],
        t.Annotated[str, p.Tag("string")],
    ],
    p.Field(discriminator=p.Discriminator(_package_string_discriminator)),
    p.BeforeValidator(lambda value: _package_name_validator(value, keep_strings=True)),
]


def _packages_discriminator(value: t.Any) -> t.Literal["list", "single"]:
    """
    Discriminate between a list of package types and a single package type.
    """
    if isinstance(value, list):
        return "list"
    return "single"


Packages = t.Annotated[
    t.Union[
        t.Annotated[PackageField, p.Tag("single")],
        t.Annotated[list[PackageField], p.Tag("list")],
    ],
    p.Field(discriminator=p.Discriminator(_packages_discriminator)),
]


class _BaseModel(p.BaseModel):
    model_config = p.ConfigDict(frozen=True, extra="allow", validate_default=True)


class SessionLint(_BaseModel):
    """
    Lint session config.
    """

    default: bool = True
    code_files: t.Union[t.Literal["default"], list[str]] = "default"
    extra_code_files: t.Optional[list[str]] = None
    module_files: t.Union[t.Literal["default"], list[str]] = "default"
    ruff_config: t.Optional[p.FilePath] = None
    ruff_modules_config: t.Optional[p.FilePath] = None
    ruff_package: Packages = PackageName(name="ruff")

    # isort:
    run_isort: bool = True
    isort_config: t.Optional[p.FilePath] = None
    isort_modules_config: t.Optional[p.FilePath] = None
    isort_package: Packages = PackageName(name="isort")

    # black:
    run_black: bool = True
    run_black_modules: t.Optional[bool] = None
    black_config: t.Optional[p.FilePath] = None
    black_modules_config: t.Optional[p.FilePath] = None
    black_package: Packages = PackageName(name="black")

    # ruff format:
    run_ruff_format: bool = False
    ruff_format_config: t.Optional[p.FilePath] = None
    ruff_format_modules_config: t.Optional[p.FilePath] = None
    ruff_format_package: t.Optional[Packages] = None

    # ruff autofix:
    run_ruff_autofix: bool = False
    ruff_autofix_config: t.Optional[p.FilePath] = None
    ruff_autofix_modules_config: t.Optional[p.FilePath] = None
    ruff_autofix_package: t.Optional[Packages] = None
    ruff_autofix_select: list[str] = []

    # ruff check:
    run_ruff_check: bool = False
    ruff_check_config: t.Optional[p.FilePath] = None
    ruff_check_modules_config: t.Optional[p.FilePath] = None
    ruff_check_package: t.Optional[Packages] = None

    # flake8:
    run_flake8: bool = True
    flake8_config: t.Optional[p.FilePath] = None
    flake8_modules_config: t.Optional[p.FilePath] = None
    flake8_package: Packages = PackageName(name="flake8")

    # pylint:
    run_pylint: bool = True
    pylint_rcfile: t.Optional[p.FilePath] = None
    pylint_modules_rcfile: t.Optional[p.FilePath] = None
    pylint_package: Packages = PackageName(name="pylint")
    pylint_ansible_core_package: t.Optional[Packages] = PackageName(name="ansible-core")
    pylint_extra_deps: list[PackageFieldOrString] = []

    # yamllint:
    run_yamllint: bool = True
    yamllint_config: t.Optional[p.FilePath] = None
    yamllint_config_plugins: t.Optional[p.FilePath] = None
    yamllint_config_plugins_examples: t.Optional[p.FilePath] = None
    yamllint_config_extra_docs: t.Optional[p.FilePath] = None
    yamllint_package: Packages = PackageName(name="yamllint")
    yamllint_antsibull_docutils_package: Packages = PackageName(
        name="antsibull-docutils"
    )

    # mypy:
    run_mypy: bool = True
    mypy_config: t.Optional[p.FilePath] = None
    mypy_modules_config: t.Optional[p.FilePath] = None
    mypy_package: Packages = PackageName(name="mypy")
    mypy_ansible_core_package: t.Optional[Packages] = PackageName(name="ansible-core")
    mypy_extra_deps: list[PackageFieldOrString] = []

    # antsibull-nox config lint:
    run_antsibullnox_config_lint: bool = True


class SessionDocsCheck(_BaseModel):
    """
    Docs check session config.
    """

    default: bool = True

    antsibull_docs_package: Packages = PackageName(name="antsibull-docs")
    ansible_core_package: Packages = PackageName(name="ansible-core")
    validate_collection_refs: t.Optional[t.Literal["self", "dependent", "all"]] = None
    extra_collections: list[CollectionName] = []

    codeblocks_restrict_types: t.Optional[list[str]] = None
    codeblocks_restrict_type_exact_case: bool = True
    codeblocks_allow_without_type: bool = True
    codeblocks_allow_literal_blocks: bool = True
    antsibull_docutils_package: Packages = PackageName(name="antsibull-docutils")


class SessionLicenseCheck(_BaseModel):
    """
    License check session config.
    """

    default: bool = True

    run_reuse: bool = True
    reuse_package: Packages = PackageName(name="reuse")
    run_license_check: bool = True
    license_check_extra_ignore_paths: list[str] = []


class ExecutionEnvironmentConfig(_BaseModel):
    """
    Execution enviroment check session config.
    """

    name: str
    description: t.Optional[str] = None

    # EE definition
    version: t.Literal[3] = 3
    base_image_name: t.Optional[str] = None  # implicit default
    ansible_core_source: t.Literal["package_pip", "package_system"] = "package_pip"
    ansible_core_package: t.Optional[str] = None
    ansible_runner_source: t.Literal["package_pip", "package_system"] = "package_pip"
    ansible_runner_package: t.Optional[str] = None
    system_packages: list[str] = []
    python_packages: list[str] = []
    python_interpreter_package: t.Optional[str] = None
    python_path: t.Optional[str] = None
    config: dict[str, t.Any] = {}

    # EE tests
    test_playbooks: list[str]
    runtime_environment: dict[str, str] = {}
    runtime_container_options: list[str] = []
    runtime_extra_vars: dict[str, str] = {}

    def to_execution_environment_config(self) -> dict[str, t.Any]:
        """
        Convert TOML config to execution environment YAML.
        """

        dependencies: dict[str, t.Any] = {}
        if self.ansible_core_package is not None:
            dependencies["ansible_core"] = {
                self.ansible_core_source: self.ansible_core_package
            }
        if self.ansible_runner_package is not None:
            dependencies["ansible_runner"] = {
                self.ansible_runner_source: self.ansible_runner_package
            }
        if self.python_interpreter_package is not None:
            python_interpreter: dict[str, t.Any] = {
                "package_system": self.python_interpreter_package
            }
            if self.python_path is not None:
                python_interpreter["python_path"] = self.python_path
            dependencies["python_interpreter"] = python_interpreter
        if self.system_packages:
            dependencies["system"] = self.system_packages
        if self.python_packages:
            dependencies["python"] = self.python_packages

        simple_config = create_ee_config(
            version=self.version,
            base_image=(
                "registry.fedoraproject.org/fedora-toolbox:latest"
                if self.base_image_name is None
                else self.base_image_name
            ),
            base_image_is_default=self.base_image_name is None,
            dependencies=dependencies,
            config=self.config,
        )

        ee_config = {**simple_config, **self.config}

        return ee_config


class SessionExecutionEnvironmentCheck(_BaseModel):
    """
    Execution environment check session.
    """

    default: bool = False
    ansible_builder_package: Packages = PackageName(name="ansible-builder")
    ansible_core_package: t.Optional[Packages] = None
    ansible_navigator_package: Packages = PackageName(name="ansible-navigator")

    execution_environments: list[ExecutionEnvironmentConfig]


class ActionGroup(_BaseModel):
    """
    Information about an action group.
    """

    # Name of the action group.
    name: str
    # Regex pattern to match modules that could belong to this action group.
    pattern: str
    # Doc fragment that members of the action group must have, but no other module
    # must have
    doc_fragment: str
    # Exclusion list of modules that match the regex, but should not be part of the
    # action group. All other modules matching the regex are assumed to be part of
    # the action group.
    exclusions: list[str] = []


class AvoidCharacterGroup(_BaseModel):
    """
    Information about characters/regexes to avoid in files.
    """

    # User-friendly name
    name: t.Optional[str] = None

    # Regular expression
    regex: str

    # Extensions, paths, and directories to look for.
    # If None (not specified), will consider all files.
    match_extensions: t.Optional[list[str]] = None
    match_paths: t.Optional[list[str]] = None
    match_directories: t.Optional[list[str]] = None

    # Extensions, paths, and directories to skip.
    skip_extensions: list[str] = []
    skip_paths: list[str] = []
    skip_directories: list[str] = []


class SessionExtraChecks(_BaseModel):
    """
    Extra checks session config.
    """

    default: bool = True

    # no-unwanted-files:
    run_no_unwanted_files: bool = True
    no_unwanted_files_module_extensions: list[str] = [".cs", ".ps1", ".psm1", ".py"]
    no_unwanted_files_other_extensions: list[str] = [".py", ".pyi"]
    no_unwanted_files_yaml_extensions: list[str] = [".yml", ".yaml"]
    no_unwanted_files_skip_paths: list[str] = []
    no_unwanted_files_skip_directories: list[str] = []
    no_unwanted_files_yaml_directories: list[str] = [
        "plugins/test/",
        "plugins/filter/",
    ]
    no_unwanted_files_allow_symlinks: bool = False

    # action-groups:
    run_action_groups: bool = False
    action_groups_config: list[ActionGroup] = []

    # no-trailing-whitespace:
    run_no_trailing_whitespace: bool = False
    no_trailing_whitespace_skip_paths: list[str] = []
    no_trailing_whitespace_skip_directories: list[str] = []

    # avoid-characters:
    run_avoid_characters: bool = False
    avoid_character_group: list[AvoidCharacterGroup] = []


class SessionBuildImportCheck(_BaseModel):
    """
    Collection build and Galaxy import session config.
    """

    default: bool = True

    ansible_core_package: Packages = PackageName(name="ansible-core")
    run_galaxy_importer: bool = True
    galaxy_importer_package: Packages = PackageName(name="galaxy-importer")
    # https://github.com/ansible/galaxy-importer#configuration
    galaxy_importer_config_path: t.Optional[p.FilePath] = None
    galaxy_importer_always_show_logs: bool = False


class DevelLikeBranch(_BaseModel):
    """
    A Git repository + branch for a devel-like branch of ansible-core.
    """

    repository: t.Optional[str] = None
    branch: str

    @p.model_validator(mode="before")
    @classmethod
    def _pre_validate(cls, values: t.Any) -> t.Any:
        if isinstance(values, str):
            return {"branch": values}
        if (
            isinstance(values, list)
            and len(values) == 2
            and all(isinstance(v, str) for v in values)
        ):
            return {"repository": values[0], "branch": values[1]}
        return values


class SessionAnsibleTestSanity(_BaseModel):
    """
    Ansible-test sanity tests session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []
    skip_tests: list[str] = []
    allow_disabled: bool = False
    enable_optional_errors: bool = False


class SessionAnsibleTestUnits(_BaseModel):
    """
    Ansible-test unit tests session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []


class AnsibleValueExplicit(p.BaseModel):
    """
    An explicit value.
    """

    type: t.Literal["value"] = "value"
    value: t.Any

    template_value: t.Optional[str] = None

    def to_utils_instance(self) -> _AnsibleValueExplicit:
        """
        Convert config object to runtime Ansible value object.
        """
        return _AnsibleValueExplicit(
            value=self.value,
            template_value=self.template_value,
        )


class AnsibleValueFromEnv(p.BaseModel):
    """
    A value taken from an environment variable
    """

    type: t.Literal["env"] = "env"
    name: str
    fallback: t.Any = None
    unset_if_not_set: bool = False

    template_value: t.Optional[str] = None

    def to_utils_instance(self) -> _AnsibleValueFromEnv:
        """
        Convert config object to runtime Ansible value object.
        """
        return _AnsibleValueFromEnv(
            name=self.name,
            fallback=self.fallback,
            unset_if_not_set=self.unset_if_not_set,
            template_value=self.template_value,
        )


AnsibleValue = t.Union[
    AnsibleValueExplicit,
    AnsibleValueFromEnv,
]


_VALID_ANSIBLE_VALUE_TYPE_NAMES = tuple(
    p.model_fields["type"].default for p in t.get_args(AnsibleValue)
)


def ansible_value_validator(value: t.Any) -> t.Any:
    """
    Convert non-dicts into AnsibleValueExplicit instances.
    """
    if isinstance(value, (str, int, float, bool)):
        return AnsibleValueExplicit(value=value)
    if isinstance(value, dict):
        # Special-casing for "type" to provide a clean error message
        if "type" not in value or value["type"] not in _VALID_ANSIBLE_VALUE_TYPE_NAMES:
            raise ValueError(
                "Must provide a valid 'type' when specifying a value."
                + f" Valid types are: {', '.join(_VALID_ANSIBLE_VALUE_TYPE_NAMES)}"
            )
    return value


AnsibleValueField = t.Annotated[
    AnsibleValue,
    p.Field(discriminator="type"),
    p.BeforeValidator(ansible_value_validator),
]


class SessionAnsibleTestIntegrationWDefaultContainer(_BaseModel):
    """
    Ansible-test integration tests with default container session config.
    """

    default: bool = False

    include_devel: bool = False
    include_milestone: bool = False
    add_devel_like_branches: list[DevelLikeBranch] = []
    min_version: t.Optional[PVersion] = None
    max_version: t.Optional[PVersion] = None
    except_versions: list[PAnsibleCoreVersion] = []
    core_python_versions: dict[t.Union[PAnsibleCoreVersion, str], list[PVersion]] = {}
    controller_python_versions_only: bool = False
    ansible_vars_from_env_vars: dict[str, str] = {}
    ansible_vars: dict[str, AnsibleValueField] = {}

    @p.model_validator(mode="after")
    def _validate_core_keys(self) -> t.Self:
        branch_names = [dlb.branch for dlb in self.add_devel_like_branches]
        for key in self.core_python_versions:
            if isinstance(key, Version) or key in {"devel", "milestone"}:
                continue
            if key in branch_names:
                continue
            raise ValueError(
                f"Unknown ansible-core version or branch name {key!r} in core_python_versions"
            )
        return self


class SessionAnsibleTestIntegrationSession(_BaseModel):
    """
    Explicit ansible-test integration test session(s).
    """

    # The following fields allow one or more value; if multiple values are
    # specified, one session is created for every combination of values
    ansible_core: t.Union[PAnsibleCoreVersion, list[PAnsibleCoreVersion]]
    docker: t.Union[str, list[str], None] = None
    remote: t.Union[str, list[str], None] = None
    python_version: t.Union[PVersion, list[PVersion], None] = None
    target: t.Union[str, list[str], None] = None
    gha_container: t.Union[str, list[str], None] = None

    # Can only be used if ansible_core == "devel"
    devel_like_branch: t.Union[DevelLikeBranch, None] = None

    ansible_vars: dict[str, AnsibleValueField] = {}
    session_name_template: t.Union[str, None] = None
    display_name_template: t.Union[str, None] = None
    description_template: t.Union[str, None] = None

    # Tags for session registration; can be used by matrix generator
    tags: list[str] = []

    @p.model_validator(mode="after")
    def _validate_after(self) -> t.Self:
        has_docker = self.docker is not None
        has_remote = self.remote is not None
        if has_docker == has_remote:
            raise ValueError("Exactly one of remote and docker must be specified")

        if self.devel_like_branch is not None and self.ansible_core != ["devel"]:
            raise ValueError(
                "devel_like_branch can only be used if ansible_core is set to 'devel'"
            )
        return self


class SessionAnsibleTestIntegrationGroup(_BaseModel):
    """
    Group of explicit ansible-test integration tests.
    """

    session_name: t.Union[str, None] = None
    description: t.Union[str, None] = None

    docker: t.Union[str, list[str], None] = None
    remote: t.Union[str, list[str], None] = None
    python_version: t.Union[PVersion, list[PVersion], None] = None
    target: t.Union[str, list[str], None] = None
    gha_container: t.Union[str, list[str], None] = None

    ansible_vars: dict[str, AnsibleValueField] = {}
    sessions: list[SessionAnsibleTestIntegrationSession] = []

    session_name_template: t.Union[str, None] = None
    display_name_template: t.Union[str, None] = None
    description_template: t.Union[str, None] = None

    # Tags for session registration; can be used by matrix generator
    tags: list[str] = []


class SessionAnsibleTestIntegration(_BaseModel):
    """
    Explicit ansible-test integration tests.
    """

    default: bool = False

    ansible_vars: dict[str, AnsibleValueField] = {}
    sessions: list[SessionAnsibleTestIntegrationSession] = []
    groups: list[SessionAnsibleTestIntegrationGroup] = []

    session_name_template: str = (
        "ansible-test-integration-{target_dash}{ansible_core}"
        "{dash_docker_short}{dash_remote}{dash_python_version}"
    )
    display_name_template: str = (
        "Ⓐ{ansible_core}{plus_py_python_version}{plus_docker_short}{plus_remote}"
    )
    description_template: str = (
        "Run integration tests with ansible-core {ansible_core}, {docker_short}{remote}"
    )

    # Tags for session registration; can be used by matrix generator
    tags: list[str] = []


class SessionAnsibleLint(_BaseModel):
    """
    Ansible-lint session config.
    """

    default: bool = True

    ansible_lint_package: Packages = PackageName(name="ansible-lint")
    additional_requirements_files: list[str] = []
    strict: bool = False


class Sessions(_BaseModel):
    """
    Configuration of nox sessions to add.
    """

    lint: t.Optional[SessionLint] = None
    docs_check: t.Optional[SessionDocsCheck] = None
    license_check: t.Optional[SessionLicenseCheck] = None
    extra_checks: t.Optional[SessionExtraChecks] = None
    build_import_check: t.Optional[SessionBuildImportCheck] = None
    ansible_test_sanity: t.Optional[SessionAnsibleTestSanity] = None
    ansible_test_units: t.Optional[SessionAnsibleTestUnits] = None
    ansible_test_integration_w_default_container: t.Optional[
        SessionAnsibleTestIntegrationWDefaultContainer
    ] = None
    ansible_test_integration: t.Optional[SessionAnsibleTestIntegration] = None
    ansible_lint: t.Optional[SessionAnsibleLint] = None
    ee_check: t.Optional[SessionExecutionEnvironmentCheck] = None


class CollectionSource(_BaseModel):
    """
    Source from which to install a collection.
    """

    source: str

    @p.model_validator(mode="before")
    @classmethod
    def _pre_validate(cls, values):
        if isinstance(values, str):
            return {"source": values}
        return values


class VCS(_BaseModel):
    """
    Version Control System configuration.
    """

    vcs: t.Literal["git"]
    development_branch: str
    # Wildcards are allowed in the lists:
    # (https://docs.python.org/3/library/fnmatch.html)
    stable_branches: list[str] = []


class CollectionConfig(_BaseModel):
    """
    Global collection configuration that isn't part of any other configuration.
    """

    min_python_version: t.Union[MinPythonVersionConstants, PVersion] = "default"


class Config(_BaseModel):
    """
    The contents of a antsibull-nox config file.
    """

    collection_sources: dict[CollectionName, CollectionSource] = {}
    collection_sources_per_ansible: dict[
        PAnsibleCoreVersion, dict[CollectionName, CollectionSource]
    ] = {}
    vcs: t.Optional[VCS] = None
    sessions: Sessions = Sessions()
    collection: CollectionConfig = CollectionConfig()


def load_config_from_toml(path: str | os.PathLike) -> Config:
    """
    Load a config TOML file.
    """
    with open(path, "rb") as f:
        try:
            data = _load_toml(f)
        except ValueError as exc:
            raise ValueError(f"Error while reading {path}: {exc}") from exc
    return Config.model_validate(data)


def lint_config_toml_messages() -> list[Message]:
    """
    Lint config files
    """
    path = CONFIG_FILENAME
    errors = []
    forbid_extras(Config)
    try:
        with open(path, "rb") as f:
            data = _load_toml(f)
        Config.model_validate(data)
    except p.ValidationError as exc:
        for error in get_formatted_error_messages(exc):
            errors.append(
                Message(
                    file=path,
                    start=None,
                    end=None,
                    level="error",
                    id=None,
                    message=f"{error}",
                )
            )
    except ValueError as exc:
        errors.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=f"{exc}",
            )
        )
    except FileNotFoundError:
        errors.append(
            Message(
                file=path,
                start=None,  # pylint: disable=R0801
                end=None,
                level="error",
                id=None,
                message="File does not exist",
            )
        )
    except IOError as exc:
        errors.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=f"{exc}",
            )
        )
    return errors


def lint_config_toml() -> list[str]:
    """
    Lint config files
    """
    return [
        f"{message.file}: {message.message}" for message in lint_config_toml_messages()
    ]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utilities.
"""

from __future__ import annotations

import fnmatch
import re
import typing as t

_NUMBER = re.compile("([0-9]+)")


class SortableBranchName:
    """
    A branch name that can be compared lexicographically,
    with special handling for integers.
    """

    def __init__(self, branch_name: str) -> None:
        self.branch_name = branch_name
        self.parts: list[
            tuple[t.Literal["i"], int, str] | tuple[t.Literal["l"], str]
        ] = []
        for index, part in enumerate(_NUMBER.split(branch_name)):
            if index % 2 == 1:
                self.parts.append(("i", int(part), part))
            else:
                for letter in part:
                    self.parts.append(("l", letter))

    def __repr__(self) -> str:
        parts = ",".join(repr(p[-1]) for p in self.parts)
        return f"SortableBranchName({self.branch_name!r}, parts={parts})"

    def __str__(self) -> str:
        return self.branch_name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SortableBranchName):
            return False
        return self.branch_name == other.branch_name

    def __lt__(self, other: SortableBranchName) -> bool:
        for a, b in zip(self.parts, other.parts):
            if a[0] == b[0]:
                if a[-1] != b[-1]:
                    # a[1] and b[1] are of the same type, but mypy doesn't notice...
                    return a[1] < b[1]  # type: ignore
            elif a[0] == "l":  # and b[0] == "i"
                # a[1] is a string, but mypy doesn't notice...
                return a[1] < "0"  # type: ignore
            else:  # a[0] == "i" and b[0] == "l"
                # b[1] is a string, but mypy doesn't notice...
                return "9" < b[1]  # type: ignore
        return len(self.parts) < len(other.parts)


def matches(branch: str, branch_list: list[str]) -> bool:
    """
    Test whether the given branch name matches one of the given patterns.
    """
    return any(
        fnmatch.fnmatch(branch, branch_pattern) for branch_pattern in branch_list
    )



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all plugin and module documentation adheres to yamllint."""

from __future__ import annotations

import ast
import io
import os
import re
import sys
import traceback
import typing as t

import yaml
from antsibull_nox_data_util import (  # type: ignore
    Level,
    Location,
    Message,
    report_result,
    setup,
)
from yamllint import linter
from yamllint.config import YamlLintConfig
from yamllint.linter import PROBLEM_LEVELS

REPORT_LEVELS: dict[PROBLEM_LEVELS, Level] = {
    "warning": "warning",
    "error": "error",
}

EXAMPLES_FMT_RE = re.compile(r"^# fmt:\s+(\S+)")

EXAMPLES_SECTION = "EXAMPLES"


def lint(
    *,
    messages: list[Message],
    path: str,
    data: str,
    row_offset: int,
    col_offset: int,
    section: str,
    config: YamlLintConfig,
    note: str | None = None,
    exact: bool = True,
) -> None:
    # If the string start with optional whitespace + linebreak, skip that line
    idx = data.find("\n")
    if idx >= 0 and (idx == 0 or data[:idx].isspace()):
        data = data[idx + 1 :]
        row_offset += 1
        col_offset = 0

    try:
        problems = linter.run(
            io.StringIO(data),
            config,
            path,
        )
        for problem in problems:
            level = REPORT_LEVELS.get(problem.level)
            if level is None:
                continue
            msg = f"{section}: {problem.level}: {problem.desc}"
            if problem.rule:
                msg += f"  ({problem.rule})"
            messages.append(
                Message(
                    file=path,
                    start=Location(
                        line=row_offset + problem.line,
                        # The col_offset is only valid for line 1; otherwise the offset is 0
                        column=(col_offset if problem.line == 1 else 0)
                        + problem.column,
                        exact=exact,
                    ),
                    end=None,
                    level=level,
                    id=None,
                    message=msg,
                    note=note,
                )
            )
    except Exception as exc:
        error = str(exc).replace("\n", " / ")
        messages.append(
            Message(
                file=path,
                start=Location(
                    line=row_offset + 1,
                    column=col_offset + 1,
                    exact=exact,
                ),
                end=None,
                level="error",
                id=None,
                message=(
                    f"{section}: Internal error while linting YAML: exception {type(exc)}:"
                    f" {error}; traceback: {traceback.format_exc()!r}"
                ),
                note=note,
            )
        )


def iterate_targets(
    assignment: ast.Assign,
) -> t.Iterable[tuple[ast.Constant, str, str]]:
    if not isinstance(assignment.value, ast.Constant):
        return
    if not isinstance(assignment.value.value, str):
        return
    for target in assignment.targets:
        try:
            yield assignment.value, assignment.value.value, target.id  # type: ignore
        except AttributeError:
            continue


def process_python_file(
    messages: list[Message],
    path: str,
    config: YamlLintConfig,
    config_examples: YamlLintConfig,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as f:
            root = ast.parse(f.read(), filename=path)
    except Exception as exc:
        messages.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=(
                    f"Error while parsing Python code: exception {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            )
        )
        return

    is_doc_fragment = path.startswith("plugins/doc_fragments/")

    # We look for top-level assignments and classes
    for child in root.body:
        if (
            is_doc_fragment
            and isinstance(child, ast.ClassDef)
            and child.name == "ModuleDocFragment"
        ):
            for fragment in child.body:
                if not isinstance(fragment, ast.Assign):
                    continue
                for constant, data, fragment_name in iterate_targets(fragment):
                    lint(
                        messages=messages,
                        path=path,
                        data=data,
                        row_offset=constant.lineno - 1,
                        col_offset=constant.col_offset - 1,
                        section=fragment_name,
                        config=config,
                    )
        if not isinstance(child, ast.Assign):
            continue
        for constant, data, section in iterate_targets(child):
            if section not in ("DOCUMENTATION", "EXAMPLES", "RETURN"):
                continue

            # Handle special values
            if data in ("#", " # ") and section == "RETURN":
                # Not skipping it here could result in all kind of linting errors,
                # like no document start, or trailing space.
                continue

            # Check for non-YAML examples
            if section == EXAMPLES_SECTION:
                fmt_match = EXAMPLES_FMT_RE.match(data.lstrip())
                if fmt_match and fmt_match.group(1) != "yaml":
                    continue

            # Parse the (remaining) string content
            lint(
                messages=messages,
                path=path,
                data=data,
                row_offset=constant.lineno - 1,
                col_offset=constant.col_offset - 1,
                section=section,
                config=config_examples if section == EXAMPLES_SECTION else config,
            )


def process_sidecar_docs_file(
    messages: list[Message],
    path: str,
    config_examples: YamlLintConfig,
) -> None:
    try:
        # TODO: get hold of YAML structure so we also get correct line/col numbers
        #       inside EXAMPLES!
        with open(path, "rb") as stream:
            root = yaml.load(stream, Loader=yaml.SafeLoader)
    except Exception as exc:
        messages.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=(
                    f"Error while parsing Python code: exception {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            )
        )
        return

    if not isinstance(root, dict):
        return
    examples = root.get(EXAMPLES_SECTION)
    if not isinstance(examples, str):
        return

    # Check for non-YAML examples
    fmt_match = EXAMPLES_FMT_RE.match(examples.lstrip())
    if fmt_match and fmt_match.group(1) != "yaml":
        return

    lint(
        messages=messages,
        path=path,
        data=examples,
        row_offset=0,  # TODO
        col_offset=0,  # TODO
        section=EXAMPLES_SECTION,
        config=config_examples,
        note="Line/column are relative to EXAMPLES string contents",
        exact=False,
    )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    config: str | None = extra_data.get("config")
    config_examples: str | None = extra_data.get("config_examples")

    if config:
        yamllint_config = YamlLintConfig(file=config)
    else:
        yamllint_config = YamlLintConfig(content="extends: default")

    if config_examples:
        yamllint_config_examples = YamlLintConfig(file=config_examples)
    else:
        yamllint_config_examples = yamllint_config

    messages: list[Message] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        if path.endswith(".py"):
            process_python_file(
                messages, path, yamllint_config, yamllint_config_examples
            )
        if path.endswith((".yml", ".yaml")):
            process_sidecar_docs_file(messages, path, yamllint_config_examples)

    return report_result(messages)


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

from pathlib import Path

import pytest

from antsibull_nox.python.imports import (
    PythonImport,
    PythonModule,
    get_all_module_data,
    get_imported_packages,
    get_module_data_from_source_root,
)

PYTHON_IMPORT_BELONGS_TO: list[tuple[PythonImport, tuple[str, ...], bool]] = [
    (PythonImport(("foo",)), ("foo", "bar"), False),
    (PythonImport(("foo",)), ("foo",), True),
    (PythonImport(("foo", "bar")), ("foo",), True),
]


@pytest.mark.parametrize(
    "python_import, module_path, expected_result",
    PYTHON_IMPORT_BELONGS_TO,
)
def test_python_import_belongs_to(
    python_import: PythonImport, module_path: tuple[str, ...], expected_result: bool
) -> None:
    assert python_import.belongs_to(module_path) == expected_result


GET_IMPORTED_PACKAGES: list[tuple[tuple[str, ...], str, set[PythonImport], bool]] = [
    (
        ("foo", "bar", "baz"),
        r"""
import baz
import baz.bam
import baz.bam.bar as fdsa
from foo.baz import bam_1
from . import bam_2 as asdf
from .bar import bam_3
from ..bar import bam_4
from ...bar import bam_5
""",
        {
            PythonImport(
                symbol_path=("baz",),
            ),
            PythonImport(
                symbol_path=(
                    "baz",
                    "bam",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "baz",
                    "bam",
                    "bar",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "foo",
                    "bar",
                    "bar",
                    "bam_3",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "foo",
                    "bar",
                    "bam_4",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "foo",
                    "bar",
                    "bam_2",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "bar",
                    "bam_5",
                ),
            ),
            PythonImport(("foo", "baz", "bam_1")),
        },
        True,
    ),
    (
        ("foo", "bar", "baz"),
        r"""
from .bam import bam_1
from ..bam import bam_2
from ...bam import bam_3
from ....bam import bam_4
from .....bam import bam_5
""",
        {
            PythonImport(
                symbol_path=(
                    "foo",
                    "bar",
                    "bam",
                    "bam_1",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "foo",
                    "bam",
                    "bam_2",
                ),
            ),
            PythonImport(
                symbol_path=(
                    "bam",
                    "bam_3",
                ),
            ),
        },
        False,
    ),
    (
        ("foo", "bar", "baz"),
        r"""
import .bam  # syntax error
from .bam import bam_1
""",
        set(),
        False,
    ),
]


@pytest.mark.parametrize(
    "module_path, content, expected_imports, expected_parsable",
    GET_IMPORTED_PACKAGES,
)
def test_get_imported_packages(
    module_path: tuple[str, ...],
    content: str,
    expected_imports: set[PythonImport],
    expected_parsable: bool,
    tmp_path: Path,
) -> None:
    path = tmp_path / "foo.py"
    path.write_text(content)
    imports, parsable = get_imported_packages(path, module_path)
    assert imports == expected_imports
    assert parsable == expected_parsable


def test_get_module_data_from_source_root(tmp_path: Path) -> None:
    dir_1 = tmp_path / "dir_1"
    dir_1.mkdir()
    file_1 = dir_1 / "file_1.py"
    file_1.write_text(r"""
import foo
from .bar import asdf
""")
    assert get_module_data_from_source_root(
        file_1, source_root=tmp_path
    ) == PythonModule(
        path=file_1,
        module_path=(
            "dir_1",
            "file_1",
        ),
        imports=frozenset(
            {
                PythonImport(
                    symbol_path=(
                        "dir_1",
                        "bar",
                        "asdf",
                    ),
                ),
                PythonImport(
                    symbol_path=("foo",),
                ),
            }
        ),
        parsable=True,
    )


def test_get_all_module_data(tmp_path: Path) -> None:
    file_1 = tmp_path / "file_1.py"
    file_1.write_text(r"""
import foo
from .bar import asdf
""")
    dir_1 = tmp_path / "dir_1"
    dir_1.mkdir()
    file_2 = dir_1 / "file_2.py"
    file_2.write_text(r"""
import foo
from .bar import asdf
""")
    file_3 = dir_1 / "__init__.py"
    file_3.write_text(r"""
import foo
from .bar import asdf
""")
    file_4 = dir_1 / "foo.pyx"
    file_4.write_text(r"""
import bam
from .bar import fdsa
""")
    result = set(get_all_module_data([(tmp_path, ("baz",))]))
    assert result == {
        PythonModule(
            path=file_1,
            module_path=(
                "baz",
                "file_1",
            ),
            imports=frozenset(
                {
                    PythonImport(
                        symbol_path=(
                            "baz",
                            "bar",
                            "asdf",
                        ),
                    ),
                    PythonImport(
                        symbol_path=("foo",),
                    ),
                }
            ),
            parsable=True,
        ),
        PythonModule(
            path=file_2,
            module_path=(
                "baz",
                "dir_1",
                "file_2",
            ),
            imports=frozenset(
                {
                    PythonImport(
                        symbol_path=("foo",),
                    ),
                    PythonImport(
                        symbol_path=(
                            "baz",
                            "dir_1",
                            "bar",
                            "asdf",
                        ),
                    ),
                }
            ),
            parsable=True,
        ),
        PythonModule(
            path=file_3,
            module_path=(
                "baz",
                "dir_1",
            ),
            imports=frozenset(
                {
                    PythonImport(
                        symbol_path=(
                            "baz",
                            "bar",
                            "asdf",
                        ),
                    ),
                    PythonImport(
                        symbol_path=("foo",),
                    ),
                }
            ),
            parsable=True,
        ),
    }

    result = set(get_all_module_data([(file_1, ("baz", "bam"))]))
    assert result == {
        PythonModule(
            path=file_1,
            module_path=(
                "baz",
                "bam",
            ),
            imports=frozenset(
                {
                    PythonImport(
                        symbol_path=(
                            "baz",
                            "bar",
                            "asdf",
                        ),
                    ),
                    PythonImport(
                        symbol_path=("foo",),
                    ),
                }
            ),
            parsable=True,
        ),
    }



# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see COPYING
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Foo documentation fragment.
"""

from __future__ import annotations


class ModuleDocFragment:  # pylint: disable=too-few-public-methods
    """
    Standard documentation fragment
    """

    DOCUMENTATION = r"""
---
options: {}
attributes:
  action_group:
    description: >-
      Use C(group/antsibull.coverage.foo) in C(module_defaults)
      to set defaults for this module.
    support: full
    membership:
      - antsibull.coverage.foo
"""
